package net.sf.javaprinciples.core;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;

import org.apache.commons.beanutils.BeanUtils;
import org.springframework.beans.BeanWrapperImpl;

/**
 * Helper class to perform Class instantiation and retrieval of attribute type
 * and name.
 *
 * @author Kay Chevalier
 */
public class ObjectClassHelper
{

    /**
     * Given an attribute name, return the corresponding attribute class from the
     * owning object.
     *
     * @param owningObject
     * @param attributeName
     * @return
     */
    public static Class<?> getAttributeTypeByName(Object owningObject, String attributeName)
    {
        BeanWrapperImpl beanWrapper = new BeanWrapperImpl(owningObject);
        return beanWrapper.getPropertyType(attributeName);
    }


    /**
     * Given an attribute type, return the corresponding attribute name from the owning
     * object with that attribute type
     * @param owningObject
     * @param attributeType
     * @return
     */
    public static String getAttributeNameByType(Object owningObject, String attributeType)
    {
        try
        {
            java.lang.Class<?> c = java.lang.Class.forName(owningObject.getClass().getName());
            Field[] declaredFields = c.getDeclaredFields();
            for (Field field : declaredFields)
            {
                if (attributeType != null && attributeType.equals(field.getType().getName()))
                     return field.getName();
            }
        }
        catch (ClassNotFoundException e)
        {
            throw new UnexpectedException("Unable to getAttributeNameByType - ClassNotFoundException :" + e.getMessage(), e.getCause());
        }

        return null;
    }

    /**
     * Instantiates and returns an object from the given String className.
     *
     * @param className - String className
     * @return  Object - the instantiated object
     */
    public static Object createObjectFromClassName(String className)
    {
        try
        {
            return java.lang.Class.forName(className).getConstructor().newInstance();
        }
        catch (InstantiationException e)
        {
            throw new UnexpectedException(e.getMessage(), e.getCause());
        }
        catch (IllegalAccessException e)
        {
            throw new UnexpectedException(e.getMessage(), e.getCause());
        }
        catch (InvocationTargetException e)
        {
            throw new UnexpectedException(e.getMessage(), e.getCause());
        }
        catch (NoSuchMethodException e)
        {
            throw new UnexpectedException(e.getMessage(), e.getCause());
        }
        catch (ClassNotFoundException e)
        {
            throw new UnexpectedException(e.getMessage(), e.getCause());
        }
    }

    public static Object cloneBean(Object bean)
    {
        try
        {
            return BeanUtils.cloneBean(bean);
        }
        catch (IllegalAccessException e)
        {
            throw new UnexpectedException("IllegalAccessException thrown whilst trying to clone bean", e);
        }
        catch (InstantiationException e)
        {
            throw new UnexpectedException("InstantiationException thrown whilst trying to clone bean", e);
        }
        catch (InvocationTargetException e)
        {
            throw new UnexpectedException("InvocationTargetException thrown whilst trying to clone bean", e);
        }
        catch (NoSuchMethodException e)
        {
            throw new UnexpectedException("NoSuchMethodException thrown whilst trying to clone bean", e);
        }
    }
}
